#Copyright STACKEO - INRIA 2020 .
class Message:

    def __init__(self, type_: str, topic: str, size: str):
        self.type = type_
        self.topic = topic
        self.size = size
        self.action = None # publish subscribe

    def set_action(self, action: str) -> None:
        self.action = action



#Copyright STACKEO - INRIA 2020 .
from abc import abstractmethod

from Stkml.Stkml import Stkml


class RegionsMap:

    def __init__(self, default_icons: str, icon_dir: str):
        self._icon_dir = icon_dir
        self.default_icons = default_icons

    @abstractmethod
    def from_stackml(self, stackml: Stkml, output: str):
        pass



#Copyright STACKEO - INRIA 2020 .
import sys
import io


class GraphvizManger:

    def __init__(self):
        self.initial_stderr = sys.stderr
        self.redirected_error = sys.stderr = io.StringIO()

    def __enter__(self):
        return self.redirected_error.getvalue()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        sys.stderr = self.initial_stderr



#Copyright STACKEO - INRIA 2020 .
from distutils import util
import os
import xml.etree.ElementTree as ET

from jinja2 import Environment, Template
from Stkml.Drawio.Link import Link
from Stkml.Drawio.Node import Node
from Stkml.Drawio.Tier import Tier
from Stkml.Stkml import Stkml


class DrawIO:

    def __init__(self, default_icons: str, icons: str):
        self.default_icons = default_icons
        self._icon_dir = os.path.join(os.getcwd(), icons)
        self.tiers = []
        self.links = []
        self.layers = {'ServicelayerElement': 0,
                       'DatalayerElement': 0,
                       'MessaginglayerElement': 0,
                       'ConnectivitylayerElement': 0,
                       'NetworklayerElement': 0,
                       'PhysicallayerElement': 0,
                       'EnergylayerElement': 0}

        self.funcs = {1: self.creat_diagram_level1, 2: self.creat_diagram_level2}

    def add_tier(self, name, type_=None) -> Tier:
        tier = Tier(name, type_)
        if len(self.tiers) > 0:
            tier.position[0] = tier.position[0] + self.tiers[-1].position[0] + self.tiers[-1].dimension[0] + 30
        self.tiers.append(tier)
        return tier

    def add_link(self, node_source, node_sink, layer=None) -> None:
        self.links.append(Link(node_source, node_sink, layer))

    def get_node(self, node_: str) -> Node:
        nod = None
        for tier in self.tiers:
            for node in tier.nodes:
                if node_ in (node.ids[0], node.name):
                    nod = node
                    break
            else:
                continue
            break
        return nod

    def generate_drawio_xml(self, template: Template, output: str):
        xml = template.render(DrawIO=self)
        with open(f'{output}.drawio', 'w+') as drawio_file:
            drawio_file.write(xml)

    def generate_drawio_diagram(self, level: int, env: Environment, output: str):
        bound = self.funcs[level].__get__(
            self, type(self))
        return bound(env, output)


    def creat_diagram_level1(self, env: Environment, output: str):
        warnings = []
        for tier in self.tiers:
            for node in tier.nodes:
                if self._icon_dir == self.default_icons:
                    _icon = tier.type
                else:
                    _icon = node.name
                node.set_image(os.path.join(self._icon_dir, f'{_icon}.png'))
                node.dimension = [70, 80]
            tier.generate_nodes_position_level1()
        self.center_tiers()
        sources = []
        sinks = []
        for link in self.links:
            if link.source.name not in sources or link.sink not in sinks and \
                link.layer == 'NetworklayerElement':# to not duplicate links
                sources.append(link.source.name)
                sinks.append(link.sink.name)
                link.source = link.source.ids[0]
                link.sink = link.sink.ids[0]
            else:
                link.display = False

        template = env.get_template('drawio/drawio.j2')
        self.generate_drawio_xml(output=output, template=template)
        return warnings

    def creat_diagram_level2(self, env: Environment, output: str):

        warnings = self.create_link_level2()
        for tier in self.tiers:
            for node in tier.nodes:
                hachting = True
                blank_layers = []
                for layer in node.layers[-1::-1]:
                    layer.update_height(self.layers[layer.type + 'layerElement'])
                    layer_len = layer.__len__()
                    if layer_len > 0:
                        hachting = False
                        blank_layers.clear()
                    elif layer_len == 0:
                        if hachting:
                            layer.hachting()
                        else:
                            blank_layers.append(layer)
                for layer in blank_layers:
                    layer.set_blank()

                node.generate_layers_position()
                node.set_dimensions()
            tier.generate_nodes_position_level2()



        self.center_tiers()
        template = env.get_template('drawio/drawio-layer.j2')
        self.generate_drawio_xml(output=output, template=template)
        return warnings

    def create_link_level2(self):
        sources = []
        sinks = []
        to_remove = []
        warnings = []
        for link in self.links:
            layer = link.layer
            if layer:
                src = link.source
                sink = link.sink
                if link.internal and src not in sources and sink not in sinks:
                    sources.append(src)
                    sinks.append(sink)
                    sink.un_stack()
                src = src.get_layer(layer)
                sink = sink.get_layer(layer)
                if src and sink:
                    link.source = src.ids[0]
                    link.sink = sink.ids[0]
                else:
                    to_remove.append(link)
                    if not src:
                        warnings.append(f'link source {link.source.name} does not have {layer} layer')
                    if not sink:
                        warnings.append(f'link sink {link.sink.name} does not have {layer} layer')
            else:
                to_remove.append(link)
                warnings.append(f'You did not specify the layer type on link between'
                                f'{link.source.name} ans {link.sink.name}')

        for i in to_remove:
            self.links.remove(i)
        return warnings

    def from_stackml(self, stackml: Stkml) -> str or None:
        for tier in stackml.tiers:
            drawio_tier = self.add_tier(name=tier.name, type_=tier.type)
            for node in tier.nodes:
                drawio_node = drawio_tier.add_node(name=node.id_, nature=node.nature)
                drawio_node.dimension = [210, 585]
                for layer_ in self.layers:
                    drawio_layer = drawio_node.add_layer(layer_.replace('layerElement', ''))
                    layer = node.get_layer_element(layer_)
                    if layer:
                        drawio_layer.hub = layer.hub and len(layer.components) > 1
                        for component in layer.components:
                            drawio_layer.add_component(name=component.id_, type_=component.component_type)
                    self.layers[layer_] = max(self.layers[layer_], drawio_layer.dimension[1])
        for link in stackml.links:
            source = self.get_node(link.source)
            sink = self.get_node(link.sink)
            if not source:
                return f'link source {link.source}  not found'
            if not sink:
                return f'link sink {link.sink}  not found'

            layer = link.layer
            self.add_link(source, sink, layer)
        return None

    def create_drawio(self, file) -> None:
        root = ET.parse(file).getroot()
        ids = {}
        for obj in root.iter('object'):
            if obj.get('Type') == 'Tier':
                tier, tier_id = self.create_tier(obj, root)
                #root.remove(obj)
                nodes = self.get_childs(root, ['Tier', tier_id], 'Node')
                for node in nodes:
                    node, node_id, ids = self.create_node(ids, node, tier)
                    #root.remove(Node)
                    layers = self.get_childs(root, [True, node_id], tag='mxCell')
                    for layer in layers:
                        ids = self.create_layer(ids, layer, node, root)

        for link in root.iter('mxCell'):
            self.create_link(ids, link, root)

    def create_link(self, ids: dict, link: ET, root: ET) -> None:
        if link.get('Type') == 'Link':
            layer_ = self.get_parent(root, link.get('source'))
            type_ = layer_.get('Type')
            if 'layerElement' in type_:
                layer_ = type_.replace('layerElement', '')
            else:
                layer_ = None
            self.add_link(ids[link.get('source')], ids[link.get('target')], layer=layer_)

    def create_layer(self, ids: dict, layer: ET, node: Node, root: ET) -> dict:
        layer = self.get_childs(root, [True, layer.get('id')], 'Layer')
        if layer:
            layer = layer[0]
            layer_id = layer.get('id')
            ids[layer_id] = node.name
            hub = bool(util.strtobool(layer.get('Hub')))
            layer_ = node.add_layer(layer.get('Type'), hub)
            layer_id = layer.find('mxCell').get('parent')
            if hub:
                layer_id = self.get_childs(root, [True, layer_id], child_type="Hub")[0].get('id')
            componenets = self.get_childs(root, [True, layer_id], child_type='Component')
            for componenet in componenets:
                name = componenet.get('label')
                for obj in root.iter('mxCell'):
                    component_id = obj.get('parent')
                    if component_id == componenet.get('id'):
                        name_ = obj.get('value')
                        if name_ != "" and name_ not in "<span>&lt;/&gt;</span>":
                            name = name_
                            break
                layer_.add_component(name, componenet.get('Type'))
        return ids

    def create_node(self, ids: dict, node: ET, tier: Tier):
        node_label = node.get('label')
        node_id = node.get('id')
        ids[node_id] = node_label
        node = tier.add_node(name=node_label, type_=node.get('Type'), nature='Hardware') # to do
        return node, node_id, ids

    def create_tier(self, obj, root):
        tier = self.add_tier(obj.get('label'))
        tier_id = obj.find('mxCell').get('parent')
        tier_id = self.get_childs(root, [False, tier_id], 'label')
        tier_id = tier_id[0].find('mxCell').get('parent')
        return tier, tier_id

    def get_childs(self, root, parent, child_type=None, tag='object'): # parent = [parent_type, parent_id]
        childs = []
        for obj in root.iter(tag):
            if child_type:
                type_ = obj.get('Type')
                if type_ and child_type in type_:
                    if parent[0]:
                        if obj.find('mxCell').get('parent') == parent[1]:
                            childs.append(obj)
                    else:
                        if obj.get('id') == parent[1]:
                            childs.append(obj)
            else:
                if parent[0]:
                    if obj.get('parent') == parent[1]:
                        childs.append(obj)
                else:
                    if obj.get('id') == parent[1]:
                        childs.append(obj)
        return childs

    def get_parent(self, root: ET, children: ET, type_='object') -> ET or None:
        parent = None
        for parent in root.iter(type_):
            if parent.get('id') == children:
                break
        return parent


    def center_tiers(self) -> None :
        min_ = 0
        width = 0
        for tier in self.tiers:
            diff = tier.position[0] - width
            if diff < 0:
                tier.position[0] += abs(diff) + 30
            width = tier.position[0] + tier.dimension[0]
            min_ = min(min_, tier.dimension[1])

        for tier in self.tiers:
            tier.position[1] -= (tier.dimension[1] - min_)/2



#Copyright STACKEO - INRIA 2020 .

from PIL import Image
from folium import Marker, features, Popup
from jinja2 import Template

class NodeMarker(Marker):

    _template = Template(u"""
           {% macro script(this, kwargs) %}
           var {{this.get_name()}} = L.marker(
               [{{this.location[0]}}, {{this.location[1]}}],
               {
                   icon: new L.Icon.Default(),
                   {%- if this.draggable %}
                   draggable: true,
                   autoPan: true,
                   {%- endif %}
                   {%- if this.props %}
                   props : {{ this.props }}
                   {%- endif %}
                   }
               )
               .addTo({{this._parent.get_name()}});
           {% endmacro %}
           """)

    def __init__(self, location, icon_dir, icon, node):
        label = node[0]
        self.quantity = node[1]
        if self.quantity > 1:
            label = str(self.quantity)+'\n'+label
        self._icon = f'{icon}.png'
        self._icon_dir = icon_dir
        img = f'{self._icon_dir}{self._icon}'
        size = Image.open(img).size
        if size[0] < size[1]:
            width = size[1]*50/size[0]
            height = 50
        else:
            height = size[0] * 50 / size[1]
            width = 50
        self.icon = features.CustomIcon(icon_image=img, icon_size=(height, width))
        super().__init__(location=location, icon=self.icon, popup=Popup(label))
        self.props = {'quantity': self.quantity}



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement import LayerElement
from Stkml.Stkml.LayerElement.ConnectivitylayerElement import ConnectivitylayerElement
from Stkml.Stkml.LayerElement.DatalayerElement import DatalayerElement
from Stkml.Stkml.LayerElement.EnergylayerElement import EnergylayerElement
from Stkml.Stkml.LayerElement.MessaginglayerElement import MessaginglayerElement
from Stkml.Stkml.LayerElement.NetworklayerElement import NetworklayerElement
from Stkml.Stkml.LayerElement.PhysicallayerElement import PhysicallayerElement
from Stkml.Stkml.LayerElement.ServicelayerElement import ServicelayerElement
from Stkml.Stkml.LayerElement.BusinesslayerElement import BusinesslayerElement


class Node:

    def __init__(self, id_: str, cardinality: int, type_: str):
        self.id_ = id_
        self.cardinality = cardinality
        self.element_types = {'BusinesslayerElement': BusinesslayerElement,
                              'ServicelayerElement': ServicelayerElement,
                              'DatalayerElement': DatalayerElement,
                              'MessaginglayerElement': MessaginglayerElement,
                              'ConnectivitylayerElement': ConnectivitylayerElement,
                              'NetworklayerElement': NetworklayerElement,
                              'PhysicallayerElement': PhysicallayerElement,
                              'EnergylayerElement': EnergylayerElement}
        self.layers_element = []
        self.type = type_
        self.nature = ''

    def add_layer_element(self, element, type_, hub) -> LayerElement:
        layer_element = self.element_types[type_](element, hub)
        self.layers_element.append(layer_element)
        return layer_element

    def get_layer_element(self, type_: str):
        layer = None
        for layer_ in self.layers_element:
            if layer_.type == type_:
                layer = layer_
                break
        return layer

    def get_components(self, layer_type: str):
        components = None
        layer_elem = self.get_layer_element(layer_type)
        if layer_elem:
            components = layer_elem.components
        return components

    def set_nature(self, nature: str) -> None:
        self.nature = nature



#Copyright STACKEO - INRIA 2020 .
class Region:

    def __init__(self, name: str, type_: str):
        self.name = name
        self.type = type_
        self.nodes = {} # {node_id : number}

    def add_node(self, node_id: str, number: int) -> None:
        self.nodes[node_id] = number



#Copyright STACKEO - INRIA 2020 .
from typing import List

from Stkml.Stkml.LayerElement.Component.DataComponent.Data import Data


class Arg:

    def __init__(self, name: str, type_: str):
        self.name = name
        self.type = type_
        self.datas: List[Data] = []


    def add_data(self, size: float, size_unit: str, type_: str, reference: str) -> None:
        self.datas.append(Data(size, size_unit, type_, reference))



#Copyright STACKEO - INRIA 2020 .

from folium import folium, PolyLine, Map
from folium.plugins import MarkerCluster
from geopy import Nominatim

from Stkml.Diagram.Graphviz.RegionDiagram import RegionDiagram
from Stkml.Diagram.Map import RegionsMap
from Stkml.Diagram.Map.NodeMarker import NodeMarker
from Stkml.Diagram.Map.RegionNotFound import RegionNotFound
from Stkml.Stkml import Stkml


class TopologyMap(RegionsMap):

    def __init__(self, default_icons: str, icon_dir: str, diagram_attr):
        self.style = diagram_attr #not used
        super().__init__(default_icons=default_icons, icon_dir=icon_dir)

    def from_stackml(self, stackml: Stkml, output: str) -> None:
        nodes = {}
        geolocator = Nominatim(user_agent="shttps://nominatim.openstreetmap.org/search/")
        topology_map = Map()  # tiles='Stamen Toner'Terrain
        marker_cluster = MarkerCluster(icon_create_function='''
                    function(cluster) {
                        var markers = cluster.getAllChildMarkers();
                        var sum = 0;
                        for (var i = 0; i < markers.length; i++) {
                            sum += markers[i].options.props.quantity;
                        }
                        return L.divIcon({
                             html: '<div style="color:#FFFFFF;text-align: center;vertical-align: middle;background-color:#006DC9;border-radius: 50% 50% 50% 50%;" <b>'+ sum + '</b></div>',
                             className: '{marker-cluster marker-cluster-small}',
                             iconSize: new L.Point(50, 50)

                        });
                    }
                ''').add_to(topology_map)
        map_bound = [[0, 0], [0, 0]]
        for region in stackml.regions:
            nodes[region.name] = {}
            location = geolocator.geocode(region.name)
            if not location:
                raise RegionNotFound(region)
            map_bound = self.get_map_bound(location, map_bound)
            lat_ = location.latitude
            long_ = location.longitude
            for node in region.nodes.items():
                nodes[region.name][node[0]] = [lat_, long_]
                if self._icon_dir != self.default_icons:
                    _icon = node[0]
                else:
                    _icon = stackml.get_node_tier(node[0]).type
                NodeMarker([lat_, long_], icon_dir=self._icon_dir, icon=_icon,
                           node=node).add_to(marker_cluster)
                lat_ += 0.1
                long_ += 0.1
        for link in self.create_regions_links(nodes, stackml.links):
            topology_map.add_child(link)
        folium.FitBounds(bounds=map_bound, padding=[10, 10]).add_to(topology_map)
        topology_map.save(f'{output}.html')

    def get_map_bound(self, location, map_bound):
        if abs(location.latitude) > abs(map_bound[0][0]):
            map_bound[0] = [location.latitude, location.longitude]
        if abs(location.longitude) > abs(map_bound[-1][-1]):
            map_bound[1] = [location.latitude, location.longitude]
        return map_bound

    def create_regions_links(self, nodes, links):
        linked = []
        lines = []
        for node in nodes:
            #region local links
            sources = []
            for link in links:
                if link.source not in sources:
                    node_source = nodes[node].get(link.source)
                    node_sink = nodes[node].get(link.sink)
                    if node_source and node_sink:
                        sources.append(link.source)
                        linked.append(link.source)
                        lines.append(PolyLine(locations=[node_source, node_sink], color='blue'))

        # links between regions
        for link in links:
            if link.source not in linked:
                linked.append(link.source)
                nodes_source = RegionDiagram.get_all_nodes(nodes, link.source)
                nodes_sink = RegionDiagram.get_all_nodes(nodes, link.sink)
                for n_src in nodes_source:
                    for n_sink in nodes_sink:
                        lines.append(PolyLine(locations=[n_src, n_sink], color='blue'))

        return lines



#Copyright STACKEO - INRIA 2020 .
from abc import abstractmethod

from Stkml.Stkml import Stkml


class Graphviz:

    def __init__(self, default_icons: str, icon_dir: str, diagram_attr: dict):
        self._icon_dir = icon_dir
        self.default_icons = default_icons
        self.diagram_attr = diagram_attr

    @abstractmethod
    def from_stackml(self, stackml: Stkml, output: str) -> str:
        return 'Ok'



#Copyright STACKEO - INRIA 2020 .
from typing import List

from Stkml.Stkml.LayerElement import LayerElement
from Stkml.Stkml.LayerElement.Component.DataComponent import DataComponent, Job


class DatalayerElement(LayerElement):

    def __init__(self, name: str, hub):
        super().__init__(name, type(self).__name__, hub)


    def set_components(self, components) -> None:
        for component in components:
            data_component = DataComponent(id_=component.get('ComponentId'), c_type=component.get('nature'),
                                           id_ref=component.get('idRef'))
            for job in (component.get('Workflow') or []):
                job_ = data_component.add_job(job.get('JobId'))
                for prec in (job.get('prec') or []):
                    job_.add_prec(prec)
                for arg in job.get('Args'):
                    arg_ = job_.add_arg(name=arg.get('name'), type_=arg.get('type'))
                    data = arg.get('data')
                    size = data.get('size')
                    arg_.add_data(size=size.get('value'), size_unit=size.get('unit'),
                                  type_=data.get('valueType'), reference=data.get('valueRef'))
            self.add_component(data_component)



#Copyright STACKEO - INRIA 2020 .
import uuid

from Stkml.Drawio import Node


class Link:

    def __init__(self, source: Node, sink: Node, layer: str):
        self.id = uuid.uuid1()
        self.source = source
        self.sink = sink
        self.layer = layer
        self.display = True
        self.color = self.get_color()
        self.internal = self.is_internal()

    def is_internal(self):
        if self.source.type == self.sink.type:
            return True
        return False

    def get_color(self) -> str:
        colors = {'Service': '#2A9187',
                      'Data': '#2A9187',
                      'Messaging': '#2A9187',
                      'Connectivity': '#5937AE',
                      'Network': '#6D41DA',
                      'Physical': '#263239',
                      'Business': '#FD5501'}
        if self.layer:
            return colors[self.layer]
        return '#000000'



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement.Component import Component


class LayerElement:

    def __init__(self, name: str, type_: str, hub):
        self.name = name
        self.type = type_
        self.function = None
        self.provider = None
        if hub == 'Hub':
            self.hub = True
        else:
            self.hub = False
        self.components = []

    def set_function(self, function: str) -> None:
        self.function = function

    def set_provider(self, provider: str) -> None:
        self.provider = provider

    def add_component(self, component) -> None:
        self.components.append(component)

    def set_components(self, components) -> None:
        for component in components:
            self.add_component(Component(id_=component.get('ComponentId'), type_=component.get('nature'),
                                         id_ref=component.get('idRef')))

    def __len__(self):
        return 1 if self.hub else len(self.components)



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement import LayerElement
from Stkml.Stkml.LayerElement.Component.MessagingComponent import MessagingComponent


class MessaginglayerElement(LayerElement):

    def __init__(self, name: str, hub):
        super().__init__(name, type(self).__name__, hub)

    def set_components(self, components) -> None:
        for component in components:
            messaging_component = MessagingComponent(id_=component.get('ComponentId'), c_type=component.get('nature'),
                                                     id_ref=component.get('idRef'))
            protocol = component.get('protocol')
            if protocol is not None: #protocol != None
                messaging_component.set_protocol(protocol.get('name') + protocol.get('version'))
            for msg in (component.get('msg') or []):
                msg_ = messaging_component.add_message(type_=msg.get('type'), topic=msg.get('topicName'),
                                                       size=msg.get('sizeUnit'))
                msg_.set_action(msg.get('action'))
            self.add_component(messaging_component)



#Copyright STACKEO - INRIA 2020 .
# pylint: disable=pointless-statement

from diagrams import Diagram, Cluster, Edge

from Stkml.Diagram.Graphviz.Node import DiagramNode
from Stkml.Diagram.Graphviz import Graphviz
from Stkml.Diagram.Graphviz.GraphvizManger import GraphvizManger
from Stkml.Stkml import Stkml


class RegionDiagram(Graphviz):

    def __init__(self, default_icons: str, icon_dir: str, diagram_attr: dict):
        super().__init__(default_icons=default_icons, icon_dir=icon_dir, diagram_attr=diagram_attr)

    def from_stackml(self, stackml: Stkml, output: str) -> str:
        nodes = {}
        with GraphvizManger() as graphviz_manager:
            with Diagram(stackml.topology.name, filename=output, show=False, graph_attr=self.diagram_attr):
                with Cluster(stackml.topology.name):
                    for region in stackml.regions:
                        nodes[region.name] = {}
                        with Cluster(region.name):
                            for node in region.nodes:
                                if self._icon_dir != self.default_icons:
                                    _icon = node
                                else:
                                    _icon = stackml.get_node_tier(node).type
                                d_node = DiagramNode(self._icon_dir, _icon, node, region.nodes[node])
                                nodes[region.name][node] = d_node
                    self.create_regions_links(nodes, stackml.links)
        return graphviz_manager

    def create_regions_links(self, nodes, links):
        linked = []
        for node in nodes:
            #region local links
            sources = []
            for link in links:
                if link.source not in sources:
                    node_source = nodes[node].get(link.source)
                    node_sink = nodes[node].get(link.sink)
                    if node_source and node_sink:
                        sources.append(link.source)
                        linked.append(link.source)
                        edge = Edge(color="black")
                        node_source >> edge << node_sink
        # links between regions
        for link in links:
            if link.source not in linked:
                linked.append(link.source)
                nodes_source = self.get_all_nodes(nodes, link.source)
                nodes_sink = self.get_all_nodes(nodes, link.sink)
                for n_src in nodes_source:
                    for n_sink in nodes_sink:
                        edge = Edge(color="black")
                        n_src >> edge << n_sink
    @staticmethod
    def get_all_nodes(nodes, name):
        nodes_ = []
        for node in nodes:
            region = nodes[node]
            for node_ in region:
                if node_ == name:
                    nodes_.append(region[node_])
        return nodes_



#Copyright STACKEO - INRIA 2020 .
from typing import List

from Stkml.Stkml.LayerElement.Component import Component
from Stkml.Stkml.LayerElement.Component.MessagingComponent.Message import Message


class MessagingComponent(Component):

    def __init__(self, id_: str, c_type: str, id_ref: str):
        super().__init__(id_=id_, type_=c_type, id_ref=id_ref)
        self.protocol = None
        self.msgs: List[Message] = []

    def set_protocol(self, protocol: str) -> None:
        self.protocol = protocol

    def add_message(self, type_: str, topic: str, size: str) -> Message:
        msg = Message(type_=type_, topic=topic, size=size)
        self.msgs.append(msg)
        return msg



#Copyright STACKEO - INRIA 2020 .
from typing import List

from Stkml.Stkml.LayerElement.Component.DataComponent.Arg import Arg


class Job:

    def __init__(self, id_: str):
        self.id = id_
        self.prec = []
        self.state = 'Running'
        self.type = None
        self.args: List[Arg] = []

    def set_state(self, state: str) -> None:
        self.state = state

    def add_prec(self, prec: str) -> None:
        self.prec.append(prec)

    def set_type(self, type_: str) -> None:
        self.type = type_

    def add_arg(self, name: str, type_: str) -> Arg:
        arg = Arg(name, type_)
        self.args.append(arg)
        return arg



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement.Component import Component


class BusinessComponent(Component):

    def __init__(self, id_: str, c_type: str, type_: str, id_ref: str):
        super().__init__(id_=id_, type_=c_type, id_ref=id_ref)
        self.type = type_
        self.period = None
        self.value = None
        self.unit = None

    def set_value(self, value: float) -> None:
        self.value = value

    def set_unit(self, unit: str):
        self.unit = unit

    def set_period(self, period: str) -> None:
        self.period = period



#Copyright STACKEO - INRIA 2020 .
class Component:

    def __init__(self, id_: str, type_: str, id_ref=None):
        self.id_ = id_
        self.id_ref = id_ref
        self.component_type = type_
        self.provider = None
        self.name = None
        self.role = None

    def set_name(self, name: str) -> None:
        self.name = name

    def set_role(self, role: str) -> None:
        self.role = role

    def set_provider(self, provider: str) -> None:
        self.provider = provider



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.Node import Node


class Tier:

    def __init__(self, name: str, type_: str) -> object:
        self.name = name
        self.type = type_
        self.nodes = []

    def add_node(self, id_: str, cardinality: int, type_: str) -> Node:
        node = Node(id_=id_, cardinality=cardinality, type_=type_)
        self.nodes.append(node)
        return node



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement import LayerElement
from Stkml.Stkml.LayerElement.Component.BusinessComponent import BusinessComponent


class BusinesslayerElement(LayerElement):

    def __init__(self, name: str, hub):
        super().__init__(name, type(self).__name__, hub)


    def set_components(self, components) -> None:
        for component in components:
            business_component = BusinessComponent(type_=component.get('type'), id_=component.get('ComponentId'),
                                                   c_type=component.get('nature'), id_ref=component.get('idRef'))
            business_component.set_value(value=component.get('value'))
            business_component.set_unit(unit=component.get('unit'))
            business_component.set_period(period=component.get('period'))
            self.add_component(business_component)



#Copyright STACKEO - INRIA 2020 .

class ModelException(Exception):

    def __init__(self, file: str = None, node_id: str = None, msg: str = None):
        super().__init__()
        self.file = file
        self.node_id = node_id
        self.msg = msg

    def no_node_model(self):
        self.msg = f'There no NodeModels on {self.file}'

    def no_modeldef(self):
        self.msg = f'There no modeldef for {self.node_id}'

    def node_not_found(self):
        self.msg = f'The {self.node_id} model does not exist on {self.file} '

    def __str__(self):
        return self.msg



#Copyright STACKEO - INRIA 2020 .
import uuid

from Stkml.Drawio.Component import Component


class Layer:
    def __init__(self, type_: str, hub: bool):
        self.type = type_
        self.hub = hub
        self.ids = []
        self.position = [10, 27, {'Service': 5, 'Data': -2,
                                  'Messaging': 13, 'Connectivity': 16,
                                  'Network': 7, 'Physical': 7,
                                  'Business': 8, 'Energy': 4}[
                                   self.type.split('layerElement')[0]]]
                                  # layer position (x,y) and label x absis
        self.colors = [{'Service': '6ED7D3', 'Data': '3CAEA3', 'Messaging': '2A9187', 'Connectivity': '5937AE',
                                  'Network': '6D41DA', 'Physical': 'B281FF', 'Business': 'F8CECC', 'Energy': 'F8CECC'}
                       [self.type.split('layerElement')[0]],
                        'none']
        self.dimension = [280, 60]
        self.components = []
        self.hatching = False
        self.generate_ids()

    def generate_ids(self) -> None:
        for _ in range(4):
            self.ids.append(uuid.uuid1())

    def add_component(self, name: str, type_: str) -> Component:
        if self.colors[0] == 'FFFFFF':
            self.colors[0] = 'D5E8D4'
        component = Component(name, type_)
        self.set_component_position(component)
        self.components.append(component)
        return component

    def set_component_position(self, component):
        if self.hub:

            component.dimension[0] = self.dimension[0] / 2 - 16
            component.position[0] += (component.dimension[0] + 10) * len(self.components) - 8
            #c.position[0] += c.dimension[0] * len(self.components) + 10
        else:
            #component.position[0] = 0
            component.position[1] += component.dimension[1] * len(self.components) + (len(self.components) + 1) * 10
            component.dimension[1] = self.dimension[1] - 25
            component.dimension[0] = self.dimension[0] - 40
            if len(self.components) > 0:
                self.dimension[1] += component.dimension[1] + 10

    def __len__(self) -> int:
        return len(self.components)

    def set_blank(self) -> None:
        self.colors[0] = 'FFFFFF'
        self.colors[1] = '999999'

    def hachting(self) -> None:
        #self.colors[0] = 'D5DDE5'
        colors = {'Service': '60cac6', 'Data': '2fa499', 'Messaging': '1e897f', 'Connectivity': '5332a9',
                                  'Network': '663bd3', 'Physical': 'a777f4', 'Business': 'F8CECC', 'Energy': 'e9c0be'}
        self.hatching = colors[self.type.split('layerElement')[0]]

    def update_height(self, new_height):
        self.dimension[1] = new_height
        if len(self.components) > 0:
            last_comp = self.components[-1]
            dim = last_comp.dimension[1] + last_comp.position[1]
            if self.dimension[1] - dim > 20:
                for comp in self.components:
                    comp.dimension[1] = self.dimension[1] - 25



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement.Component import Component


class NetworkComponent(Component):

    def __init__(self, id_: str, c_type: str, id_ref: str):
        super().__init__(id_=id_, type_=c_type, id_ref=id_ref)
        self.value = None



#Copyright STACKEO - INRIA 2020 .
from diagrams import Node

class DiagramNode(Node):

    def __init__(self, icon_dir, icon, label, quantity=0):
        self._icon_dir = icon_dir
        self._icon = f'{icon}.png'
        self.label = label
        if quantity and quantity > 1:
            self.label = f'{quantity}\n{self.label}'
        super().__init__(self.label)



#Copyright STACKEO - INRIA 2020 .
from typing import List

from Stkml.Stkml.LayerElement.Component import Component
from Stkml.Stkml.LayerElement.Component.DataComponent.Job import Job


class DataComponent(Component):

    def __init__(self, id_: str, c_type: str, id_ref: str):
        super().__init__(id_=id_, type_=c_type, id_ref=id_ref)
        self.workflow: List[Job] = []


    def add_job(self, id_: str) -> Job:
        job = Job(id_)
        self.workflow.append(job)
        return job



#Copyright STACKEO - INRIA 2020 .
from typing import List


import jsonschema
from jsonschema import Draft7Validator

import yaml
from yaml.composer import Composer
from yaml.constructor import Constructor

class StkmlSyntaxicError(Exception):

    def __init__(self, schema: dict, file: dict, file_name: str):
        super().__init__()
        self.schema = schema
        self.file = file
        self.file_name = file_name
        self.errors_list: List[jsonschema.exceptions.ValidationError] = []


    def get_all_errors(self):
        validator = Draft7Validator(self.schema)
        errors = sorted(validator.iter_errors(self.file), key=lambda e: e.path)

        not_error_list = []
        for error in errors:
            self.get_errors_list(error, not_error_list=not_error_list)


    def get_errors_list(self, error, not_error_list):
        if len(error.context) == 0 and error.instance not in not_error_list and error not in self.errors_list:
            self.errors_list.append(error)
            return
        not_error_list.append(error.instance)
        for er_ in error.context:
            self.get_errors_list(er_, not_error_list)


    def get_path(self, error: jsonschema.exceptions.ValidationError, yaml_line_nb: dict) -> (str, int):
        error = list(error.absolute_path)
        path = error[0]
        line = yaml_line_nb[path]
        for path_ in error[1:]:
            line_ = line[path_]
            if isinstance(line_, (list, dict)):
                line = line_
            if isinstance(path_, int):
                path += '[' + str(path_) + ']'
            else:
                path += '["' + path_ + '"]'
        if isinstance(line, list):
            line = line[0]
        return path, line['__line__']

    def __str__(self) -> str:
        self.get_all_errors()
        yaml_nb = self.create_yaml_with_line_number()
        msg = f'{len(self.errors_list)} error(s) in {self.file_name}\n'
        for error in self.errors_list:
            _, line = self.get_path(error, yaml_nb)# path
            msg += f'{error.message} in "{self.file_name}", line {line}\n' #on {path}
        return msg

    def create_yaml_with_line_number(self):
        loader = yaml.Loader(open(self.file_name).read())

        def compose_node(parent, index):
            # the line number where the previous token has ended (plus empty lines)
            line = loader.line
            node = Composer.compose_node(loader, parent, index)
            node.__line__ = line + 1
            return node

        def construct_mapping(node, deep=False):
            mapping = Constructor.construct_mapping(loader, node, deep=deep)
            mapping['__line__'] = node.__line__
            return mapping

        loader.compose_node = compose_node
        loader.construct_mapping = construct_mapping
        data = loader.get_single_data()
        return data



#Copyright STACKEO - INRIA 2020 .
import uuid


class Component:

    def __init__(self, name: str, type_: str):
        self.id = uuid.uuid1()
        self.name = name
        self.type = type_
        self.position = [10, 7]
        self.dimension = [240, 35] #90



#Copyright STACKEO - INRIA 2020 .
import uuid

from Stkml.Drawio.Node import Node


class Tier:

    def __init__(self, name, type_):
        self.name = name
        self.ids = []
        self.nodes = []
        self.position = [20, 120]
        self.dimension = [300, 0]
        self.type = type_
        self.generate_ids()

    def generate_ids(self):
        for _ in range(4):
            self.ids.append(uuid.uuid1())

    def add_node(self, name: str, nature, type_=None) -> Node:
        if not self.type:
            self.type = type_.split('Node')[0]
        node = Node(name, self.type + 'Node', nature)

        self.nodes.append(node)
        return node

    def generate_nodes_position_level2(self) -> None:
        i = 0
        height = 0
        width = 0
        for node in self.nodes:
            if node.stack:
                node.position.append(int((self.dimension[0] - node.dimension[0]) / 2))
                y_axis = self.position[1] - 40 + i * 20 + height
                node.position.append(y_axis)
                height += node.dimension[1]
                width = node.dimension[0]
                self.dimension[1] += node.dimension[1] + 20


            else:
                node.position.append(width + 20)
                y_axis = self.position[1] - 40
                node.position.append(y_axis)
                width += node.dimension[0]
                height = node.dimension[1]
                self.dimension[0] += node.dimension[0] + 20 #* (i - 1)
            i += 1

        #self.dimension[1] = height + 95
        #self.dimension[0] = width + 20 * (i - 1)

    def generate_nodes_position_level1(self) -> None:
        i = 0
        dim = 0
        for node in self.nodes:
            node.position.append(int((self.dimension[0] - node.dimension[0]) / 2))
            y_axis = self.position[1] - 40 + i * 40 + dim
            node.position.append(y_axis)
            dim += node.dimension[1]
            i += 1
        self.dimension[1] = dim + i * 40 + 95



#Copyright STACKEO - INRIA 2020 .
from diagrams import Node

class Diagramlayer(Node):

    def __init__(self, label: str, height: int = 1):
        self.style = None
        self.shape = "plaintext"
        self.width = 2.5
        self.height = height
        if label == '':
            self.style = 'invis'
        super().__init__(label)
        padding = 0.1 * self.height

        self.attrs = {
            "shape": self.shape,
            "style": self.style,
            "width": str(self.width),
            "height": str(padding)
        }
        self.set_format_rules_on_parent()

    def set_format_rules_on_parent(self):
        if self._cluster:
            self._cluster.node(self._id, self.label, **self.attrs)
        else:
            self._diagram.node(self._id, self.label, **self.attrs)



#Copyright STACKEO - INRIA 2020 .
# pylint: disable=too-many-public-methods
import os
from typing import List
import copy
import json

import requests
from jinja2 import Environment, Template


import jsonschema
from jsonschema import validate
import yaml
from yaml.scanner import ScannerError

from Stkml import STACKML_URL_SCHEMA, STKML_PACKAGES, STKML_EXTENSION
from Stkml.Stkml.LayerElement import Component
from Stkml.Stkml.ModelException import ModelException

from Stkml.Stkml.Region import Region
from Stkml.Stkml.StkmlSyntaxicError import StkmlSyntaxicError
from Stkml.Stkml.StkmlSyntaxicErrorList import StkmlSyntaxicErrorList
from Stkml.Stkml.Tier import Tier
from Stkml.Stkml.Link import Link
from Stkml.Stkml.Node import Node
from Stkml.Stkml.Topology import Topology
from Stkml.Stkml.WarningList import WarningList


class Stkml:

    def __init__(self, folder=None):
        self.stackmlfile = None
        if folder:
            self.stackmlfile = 'main'
            self.stkml_folder = folder
        self.tiers: List[Tier] = []
        self.links: List[Link] = []
        self.regions: List[Region] = []
        self.topology = Topology()
        self.schema = self.get_stackml_schema()
        self.errors = []
        self.libraries = {}

    def check_file(self, file) -> dict or ScannerError:
        if not file['ignore-check']:
            if file['file']:
                with open(file['path'], 'r') as stkml_file:
                    yaml.load(stkml_file, Loader=yaml.FullLoader)
            else:
                for root, _, _file in os.walk(file['path']):
                    for file_ in _file:
                        if '.stkml.yaml' in file_:
                            with open(os.path.join(root, file_), 'r') as stkml_file:
                                yaml.load(stkml_file, Loader=yaml.FullLoader)
        return True

    def get_library(self, name):#-> FileNotFoundError or str
        library = {'path': '', 'file': True, 'ignore-check': False}
        local_file = os.path.join(self.stkml_folder, f'{name}.{STKML_EXTENSION}')
        if os.path.isfile(local_file):
            library['path'] = local_file
        else:
            local_folder = os.path.join(self.stkml_folder, name)
            if os.path.isdir(local_folder):
                library['path'] = local_folder
                library['file'] = False
            else:
                global_file = os.path.join(self.stkml_folder, '..', STKML_PACKAGES, f'{name}.{STKML_EXTENSION}')
                if os.path.isfile(global_file):
                    self.libraries[name] = global_file
                    library['path'] = global_file
                    library['ignore-check'] = True
                else:
                    global_folder = os.path.join(self.stkml_folder, '..', STKML_PACKAGES, name)
                    if os.path.isdir(global_folder):
                        library['path'] = global_folder
                        library['file'] = False
                        library['ignore-check'] = True
                    else:
                        raise FileNotFoundError(
                            f'\nThe {name.split("/")[-1].split(".")[0]} library does not exist')
        return library

    def get_stackml_schema(self) -> dict or None:
        try:
            schema = None
            request = requests.get(STACKML_URL_SCHEMA)
        except requests.exceptions.ConnectionError:
            pass
        else:
            if request.status_code == 200:
                schema = json.loads(request.content.decode("utf-8"))
            request.close()
            return schema
    def validate_schema(self, file) -> bool or Exception:
        if not file['ignore-check']:
            if file['file']:
                self._validate(file['path'])
            else:
                for root, _, _file in os.walk(file['path']):
                    for file_ in _file:
                        if '.stkml.yaml' in file_:
                            self._validate(os.path.join(root, file_))

    def _validate(self, stkml_file):
        with open(stkml_file, 'r') as stkml_:
            stkml = yaml.load(stkml_, Loader=yaml.FullLoader)
            try:
                validate(stkml, self.schema)
            except jsonschema.exceptions.ValidationError:
                self.errors.append(StkmlSyntaxicError(self.schema, stkml, stkml_file))


    def check_stackml_project(self) -> str or ScannerError:
        lib = self.get_library(self.stackmlfile)
        self.libraries[self.stackmlfile] = lib
        self.check_file(lib)
        i = 0
        with open(lib['path'], 'r') as stkml_file:
            stkml = yaml.load(stkml_file, Loader=yaml.FullLoader)
            for import_ in (stkml.get('import') or []):
                lib = self.get_library(import_)
                self.libraries[import_] = lib
                self.check_file(lib)
                i += 1
        return True

    def check_stckml_package(self) -> str or ScannerError:
        lib = self.get_library(self.stkml_folder)
        self.check_file(lib)
        self.libraries[self.stkml_folder] = lib
        return True

    def validate_stackml_project(self) -> str or StkmlSyntaxicErrorList:
        if self.schema:
            for lib in self.libraries:
                self.validate_schema(self.libraries[lib])
            if len(self.errors) > 0:
                raise StkmlSyntaxicErrorList(self.errors)
            return True
        raise Exception('Stackml schema does not found please verify your connection')

    def add_tier(self, tier_type: str) -> Tier:
        for tier in self.tiers:
            if tier.type == tier_type:
                return tier
        tier = Tier(name=tier_type, type_=tier_type)
        self.tiers.append(tier)
        return tier

    def create_architecture(self):
        with open(self.libraries.get(self.stackmlfile)["path"], 'r') as stackml_file:
            stackml = yaml.load(stackml_file, Loader=yaml.FullLoader)
            topology = stackml['topology']
            self.topology.name = topology.get('name')
            self.topology.usecase = topology.get('useCase')
            self.topology.type = topology.get('type')
            nodes = topology.get('Nodes')
            if nodes:
                self.create_tiers(nodes)
            links = topology.get('Links')
            if links:
                self.create_links(links)
            regions = topology.get('Regions')
            if regions:
                self.create_regions(regions)


    def create_regions(self, regions) -> None:
        for region in regions:
            region_ = Region(name=region.get('name'), type_=region.get('type'))
            for node in region.get('NodePopulation'):
                region_.add_node(node_id=node.get('nodeId'), number=node.get('number'))
            self.regions.append(region_)

    def create_links(self, links) -> None:
        for link in links:
            source = link.get('source')
            sink = link.get('sink')
            self.links.append(Link(source=source, sink=sink,
                                   layer=link.get('layer'), secure=link.get('secure')))
    def get_node_model(self, model_name):
        node = None
        if '.' in model_name:#extrenel model

            model = model_name.split('.')
            file = self.libraries.get(model[0])['path']
            if file:
                with open(file, 'r') as s_file:
                    model_def = yaml.load(s_file, Loader=yaml.FullLoader)
                node_id = model[1]
            else:
                raise ModelException(msg=f'{model[0]} is not imported')
        else:#internal model
            file = self.libraries.get(self.stackmlfile)['path']
            with open(file, 'r') as s_file:
                model_def = yaml.load(s_file, Loader=yaml.FullLoader)
            node_id = model_name
        model_def = model_def.get('modeldef')
        if model_def:
            model_def = model_def.get('NodeModels')
            if model_def:
                for node_def in model_def:
                    if node_def.get('name') == node_id:
                        node = node_def
                        break
            else:
                raise ModelException(msg=f'The are no NodeModels on {file}')
        else:
            raise ModelException(msg=f'The are no modeldef for {node_id}')
        if not node:
            raise ModelException(msg=f'The {node_id} model does not exist on {file}')
        return node


    def create_tiers(self, nodes) -> None:
        #warnings = []
        for node in nodes:
            model = node.get('model')
            node_def = self.get_node_model(model)
            node_type = node_def.get('technology')
            tier = node.get('tier')
            #if tier not in node_type:
            #    warnings.append(f"you are using {node_type} in {tier} Tier")
            tier = self.add_tier(tier)
            node_ = tier.add_node(id_=node.get('id'), type_=node_type,
                                  cardinality=node.get('cardinality'))
            node_.set_nature(node_def.get('nature'))
            for layer in node_.element_types:
                elem = node_def.get(layer)
                if elem:
                    hub = elem[0].get('ElementType')
                    if not hub:
                        hub = 'EndPoint'
                    else:
                        elem = elem[1:]
                    layer_elem = node_.add_layer_element(layer, type_=layer, hub=hub)
                    layer_elem.set_components(elem)
        self.sort_tiers()
        #return warnings

    def generate_files(self, environment: Environment) -> Template:
        template = environment.get_template('StkmlIn/StkMl.yaml.j2')
        return template.render(Stackml=self)

    def get_node(self, id_: str) -> Node:
        node = None
        for tier in self.tiers:
            for node_ in tier.nodes:
                if node_.id_ == id_:
                    node = node_
                    break
        return node


    def get_node_tier(self, id_: str) -> Tier or None:
        tier = None
        for tier_ in self.tiers:
            for node in tier_.nodes:
                if node.id_ == id_:
                    tier = tier_
                    break
        return tier

    def get_tier(self, tier_name: str) -> Tier or None:
        tier = None
        for tier_ in self.tiers:
            if tier_.name == tier_name:
                tier = tier_
                break
        return tier

    def get_component(self, layer, cmp_id) -> Component or None:
        comp = None
        for tier in self.tiers:
            for node in tier.nodes:
                cmps = node.get_components(layer)
                for component in (cmps or []):
                    if component.id_ == cmp_id:
                        comp = component
                        break
                else:
                    continue
                break
        return comp

    def creat_referenced_components(self) -> None:
        for i_t in range(len(self.tiers)):
            tier = self.tiers[i_t]
            for i_n in range(len(tier.nodes)):
                node = tier.nodes[i_n]
                for i_layer in range(len(node.layers_element)):
                    layer = node.layers_element[i_layer]
                    for i_comp in range(len(layer.components)):
                        comp = layer.components[i_comp]
                        id_ = comp.id_ref
                        if id_:

                            comp = self.get_component(layer.type, id_)
                            id_ = comp.id_
                            comp = copy.deepcopy(comp)
                            layer.components[i_comp] = comp
                            layer.components[i_comp].id_ = id_


    def from_drawio(self, drawio) -> None:
        for tier in drawio.tiers:
            stackml_tier = Tier(name=tier.name, type_=tier.type)
            self.tiers.append(stackml_tier)
            for node in tier.nodes:
                stackml_node = stackml_tier.add_node(id_=node.name, cardinality=1, type_=node.type)
                for layer in reversed(node.layers):
                    if len(layer.components) > 0:
                        stackml_layer = stackml_node.add_layer_element(element='', type_=layer.type, hub=layer.hub)
                        for component in layer.components:
                            stackml_layer.add_component(Component(id_=component.name,
                                                                  type_=component.type.split('Component')[0]))

        for link in drawio.links:
            source = drawio.get_node(link.source)
            sink = drawio.get_node(link.source)
            stackml_link = Link(source=source.name,
                                sink=sink.name, layer=link.Layer)
            self.links.append(stackml_link)

    def sort_tiers(self):
        tiers_name = ['Thing', 'Edge', 'Network', 'Platform', 'Application']
        sorted_tiers = []
        for tier in tiers_name:
            tier_ = self.get_tier(tier)
            if tier_:
                sorted_tiers.append(tier_)

        self.tiers = sorted_tiers



#Copyright STACKEO - INRIA 2020 .
import os

from Stkml import MODULE_DIR
from Stkml.Diagram.Graphviz.LayerDiagram import LayerDiagram
from Stkml.Diagram.Graphviz.RegionDiagram import RegionDiagram
from Stkml.Diagram.Graphviz.SystemDiagram import SystemDiagram
from Stkml.Diagram.Graphviz.Node import DiagramNode
from Stkml.Diagram.Graphviz.Layer import Diagramlayer
from Stkml.Diagram.Map.NodeMarker import NodeMarker
from Stkml.Diagram.Map.TopologyMap import TopologyMap
from Stkml.Stkml import Stkml


class StkmlDiagram:

    def __init__(self, icons: str):
        self._icon_dir = os.path.join(os.getcwd(), icons)
        self.default_icons = os.path.join(MODULE_DIR, 'templates/diagram/resources/')
        self.diagram_attr = {"pad": "0.2", "splines": "ortho", "nodesep": "0.6",
                             "ranksep": "1", "fontname": "Sans-Serif", "fontsize": "15", "fontcolor": "#2D3436"}

        self.funcs = {1: SystemDiagram, 2: RegionDiagram, 3: TopologyMap}

    def diagram_from_stackml(self, type_: int, stackml: Stkml, output: str) -> str:
        bound = self.funcs[type_](self.default_icons, self._icon_dir, self.diagram_attr)
        return bound.from_stackml(stackml, output)



#Copyright STACKEO - INRIA 2020 .
import os
import uuid
import base64

import mimetypes

from Stkml.Drawio.Layer import Layer


class Node:

    def __init__(self, name, type_, nature):
        self.name = name
        self.ids = []
        self.position = []
        self.dimension = [0, 0]
        self.image = None
        self.type = type_
        self.layers = []
        self.generate_ids()
        self.box_style = self.set_box_style(nature)
        self.stack = True

    def generate_ids(self):
        for _ in range(3):
            self.ids.append(uuid.uuid1())

    def set_image(self, image):
        """Convert a file (specified by a path) into a data URI."""
        if not os.path.exists(image):
            raise FileNotFoundError
        mime, _ = mimetypes.guess_type(image)
        with open(image, 'rb') as img:
            data = img.read()
            data64 = base64.b64encode(data)
            self.image = f'data:{mime},{data64.decode()}'

    def get_layer(self, layer: str) -> Layer:
        layer_ = None
        for node_layer in self.layers:
            if node_layer.type == layer:
                layer_ = node_layer
                break
        return layer_

    def add_layer(self, type_: str, hub=False) -> Layer:
        layer = Layer(type_, hub)
        self.layers.append(layer)
        return layer

    def generate_layers_position(self):
        for i_l in range(1, len(self.layers)):
            self.layers[i_l].position[1] = self.layers[i_l - 1].position[1] + self.layers[i_l - 1].dimension[1] + 10

    def set_dimensions(self):
        self.dimension[0] = 300 #self.layers[-1].dimension[0] + 20
        self.dimension[1] = self.layers[-1].position[1] + self.layers[-1].dimension[1] + 10

    def set_box_style(self, nature) -> str:
        box_style = ''
        if nature != 'Hardware':
            box_style = "dashed=1;"
        if nature == 'Service':

            box_style = f'{box_style}dashPattern=1 4;'
        return box_style

    def un_stack(self):
        self.stack = False



#Copyright STACKEO - INRIA 2020 .
# pylint: disable=pointless-statement

from diagrams import Diagram, Cluster, Edge

from Stkml.Diagram.Graphviz.Layer import Diagramlayer
from Stkml.Diagram.Graphviz import Graphviz
from Stkml.Diagram.Graphviz.GraphvizManger import GraphvizManger
from Stkml.Stkml import Stkml, Link


class LayerDiagram(Graphviz):

    def __init__(self, default_icons: str, icon_dir: str, diagram_attr: dict):
        super().__init__(default_icons=default_icons, icon_dir=icon_dir, diagram_attr=diagram_attr)
        self.layer_types = {'ServicelayerElement': 'Service',
                            'DatalayerElement': 'Data',
                            'MessaginglayerElement': 'Messaging',
                            'ConnectivitylayerElement': 'Connectivity',
                            'NetworklayerElement': 'Network',
                            'PhysicallayerElement': 'Physical',
                            'BusinesslayerElement': 'Business'
                            }
        self.layer_style = {"style": "dashed", "bgcolor": "#A8EEEB", "labeljust": "r", "fontsize": "9", "margin": "15"}

    def from_stackml(self, stackml: Stkml, output: str) -> str:
        layer_max_nb_cmp = []
        for layer_type in self.layer_types:
            layer_max_nb_cmp.append(self.get_layer_max_nb_components(stackml, layer_type))
        layers = {}  # { Node:{layers} }
        with GraphvizManger() as graohviz_manager:
            with Diagram(stackml.topology.name, filename=output, show=False, graph_attr=self.diagram_attr):
                with Cluster(stackml.topology.name, graph_attr={"style": "", "bgcolor": "White"}):
                    for tier in stackml.tiers:
                        with Cluster(tier.name, graph_attr={"style": "", "bgcolor": "White", "labeljust": "r"}):
                            node = tier.nodes[0]
                            layers[node.id_] = {}
                            with Cluster(node.id_, graph_attr={"style": "", "bgcolor": "#6ED7D3"}):
                                for layer_, nb_cmp in zip(self.layer_types, layer_max_nb_cmp):
                                    layer = node.get_layer_element(layer_)
                                    layer_name = f'<<I> {self.layer_types[layer_]} </I>>'
                                    d_layer = self.create_layer(layer, layer_name, nb_cmp)
                                    if d_layer:
                                        layers[node.id_][self.layer_types[layer_]] = d_layer
                    for link in stackml.links:
                        self.create_layer_link(layers, link)
        return graohviz_manager

    def get_layer_max_nb_components(self, stackml, layer) -> int:
        max_nb = 0
        tiers = stackml.tiers
        for tier in tiers:
            for node in tier.nodes:
                layer_elem = node.get_layer_element(layer)
                if layer_elem:
                    max_nb = max(max_nb, len(layer_elem))
        return max_nb

    def create_layer_link(self, layers: dict, link: Link) -> None:
        layer_src = layers.get(link.source)
        layer_sink = layers.get(link.sink)
        if layer_src and layer_sink:
            layer_src = layer_src.get(link.layer)
            layer_sink = layer_sink.get(link.layer)
            if layer_src and layer_sink:
                if link.secure:

                    edge = Edge(color="darkgreen",
                                label=(f'<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0"><TR><TD>'
                                       f'<IMG SRC="{self.default_icons}icons/secure.png"/>'
                                       '</TD></TR></TABLE>>'))  # label='Secure'
                else:
                    edge = Edge(color="firebrick")
                layer_src >> edge << layer_sink

    def create_layer(self, layer, layer_name, nb_cmp):
        layer_style = self.layer_style.copy()
        if 'Business' in layer_name:
            layer_style['bgcolor'] = '#FAC1BD'
        elif not layer:
            layer_style['bgcolor'] = 'white'
        if layer:
            with Cluster(layer_name, graph_attr=layer_style):
                width = len(layer.components) / len(layer)
                cmp_name = '<<TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">'
                if width == 1:  # EndPOint : stack of components
                    cmp_label = ['<TR>', '</TR>']
                else:
                    cmp_label = ['', '']
                    cmp_name = f'{cmp_name} <TR>'
                for component in layer.components:
                    cmp_name = (f'{cmp_name}{cmp_label[0]}<TD PORT="{component.id_}">'
                                f'{component.id_} </TD>{cmp_label[1]}')
                if width != 1:
                    cmp_name = f'{cmp_name}</TR>'
                cmp_name = f'{cmp_name}</TABLE>>'
                return Diagramlayer(label=cmp_name, height=nb_cmp)

        with Cluster(layer_name, graph_attr=layer_style):
            return  Diagramlayer(label='')



#Copyright STACKEO - INRIA 2020 .
from typing import List

from Stkml.Stkml import StkmlSyntaxicError


class StkmlSyntaxicErrorList(Exception):

    def __init__(self, errors: List[StkmlSyntaxicError]):
        super().__init__()
        self.errors = errors

    def __str__(self):
        errors_str = ''
        for error in self.errors:
            errors_str = f'{errors_str}{error.__str__()}'
        return errors_str



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml.LayerElement import LayerElement


class PhysicallayerElement(LayerElement):

    def __init__(self, name: str, hub):
        super().__init__(name, type(self).__name__, hub)



#Copyright STACKEO - INRIA 2020 .
from typing import List

class WarningList(Exception):

    def __init__(self, warnings: List[str]):
        super().__init__()
        self.warnings = warnings

    def __str__(self):
        warnings = ''
        for warn in self.warnings:
            warnings = f'{warnings}{warn}\n'
        return warnings



#Copyright STACKEO - INRIA 2020 .
from Stkml.Stkml import Region


class RegionNotFound(Exception):

    def __init__(self, region: Region):
        super().__init__()
        self.region = region


    def __str__(self):
        return f'{self.region.name} region does not found'



#Copyright STACKEO - INRIA 2020 .
class Link:
    def __init__(self, source: str, sink: str, layer: str,secure=False):
        self.source = source
        self.sink = sink
        self.layer = layer
        self.secure = secure

    def set_secure(self, secure: bool) -> None:
        self.secure = secure

    def is_internal(self, tier_src: str, tier_sink: str) -> bool:
        if tier_sink == tier_src:
            return True
        return False



#Copyright STACKEO - INRIA 2020
#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup
from setuptools import find_packages

import codecs
import os.path

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name='stkml',
    version=get_version("Stkml/__init__.py"),
    packages=find_packages(exclude=["Stkml_tests"]),
    license='GNU Affero General Public License v3',
    author="Fraoui Zakaria",
    author_email="zakaria@stackeo.io",
    description="stkml is a transpiler for stckml language to generate output for different backend as diagram, drawio, map.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://stkml.stackeo.io/",
    entry_points={
        'console_scripts': [
            'stkml = Stkml.Cli:cli',
        ],
    },
    install_requires=[
        'click',
        'decorator',
        'diagrams',
        'folium',
        'geopy',
        'Jinja2',
        'jsonschema',
        'Pillow',
        'pyyaml'
    ],
    extras_require={
        'dev': [
            'pylint',
            'coverage',
            'tox',
            'twine'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Environment :: Console"
    ],
    include_package_data=True,
    package_data={
        'templates': ['Stackml/templates/*'],
    },

)



#Copyright STACKEO - INRIA 2020
# pylint: disable=redefined-builtin

from __future__ import print_function

import json
import os
import sys
import shutil

import click
import requests

from jinja2 import Environment, FileSystemLoader
from yaml.scanner import ScannerError

from Stkml.Diagram.Map.RegionNotFound import RegionNotFound
from Stkml.Drawio import DrawIO
from Stkml.Stkml import Stkml, StkmlSyntaxicErrorList, ModelException, WarningList
from Stkml.Diagram import  StkmlDiagram
from Stkml import MODULE_DIR, __version__
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


CONTEXT_SETTINGS = {'help_option_names':['-h', '--help']}
ICONS = os.path.join(MODULE_DIR, 'templates/diagram/resources/')
@click.group()
@click.version_option(__version__)
def cli():
    latest_version = get_stackml_latest_version()
    if latest_version:
        if __version__ != latest_version:
            click.secho(f"""[WARN] deprecated stackml@{__version__}: stackml@<{latest_version}\
is no longer maintained. Upgrade to stackml@^{latest_version} using\
"pip install --upgrade stkml" """,
                        fg='yellow')

@click.command('check', help='check stackml project', context_settings=CONTEXT_SETTINGS)
@click.option('--input', '-i', 'stackmlfile', help='the stackml project to check')
@click.option('--disable-syntax-verification', '-d', help='ignore syntax verification', is_flag=True)
def check(stackmlfile: str, disable_syntax_verification: bool) -> None:
    stackml, _ = create_stackml(stackmlfile)
    try:
        check_stackml(stackml, disable_syntax_verification)
    except click.ClickException as error:
        click.secho(f"[ERROR] {error}", fg='red')
        sys.exit(1)
    else:
        click.secho("[SUCCESS] File checked successfully", fg='green')


@click.group('compile', help='compile Stackml file using a specific backend',
             context_settings=CONTEXT_SETTINGS)
@click.option('--input', '-i', 'stackmlfile', help='the file to compile')
@click.option('--disable-syntax-verification', '-d', help='ignore syntax verification', is_flag=True)
@click.pass_context
def compile(ctx: click.Context, stackmlfile: str, disable_syntax_verification: bool) -> None:
    if stackmlfile:
        stackml, working_dir = create_stackml(stackmlfile)
        try:
            check_stackml(stackml, disable_syntax_verification)
        except click.ClickException as error:
            click.secho(f"[ERROR] {error}", fg='red')
            sys.exit(1)
        else:
            try:
                stackml.create_architecture()
            except ModelException as error:
                click.secho(f"[ERROR] {error} ", fg='red')
                sys.exit(1)
            else:
                ctx.ensure_object(dict)
                ctx.obj['Stackml'] = stackml
                ctx.obj['working_dir'] = working_dir

@click.command('diagram', short_help='compile Stackml file for diagram', context_settings=CONTEXT_SETTINGS)
@click.option('--type', '-t', 'type', help='the diagram type {Architecture ,Topology or Map}',
              required=True, type=click.Choice(['1', '2', '3']))
@click.option('--icons', '-i', 'icons', default=ICONS, help='the icon folder {it contains [Node_Id].png files}')
@click.option('--output', '-o', 'outputfile', required=True, help='the output file')
@click.pass_context
def diagram(ctx: click.Context, type: str, icons: str, outputfile: str) -> None:
    """
    compile Stackml file for diagram

    Example:
    \b
        Stackml compile -i architecture.stkml diagram -t 1 -o img
    """
    if which('dot'):
        stackml = ctx.obj['Stackml']
        working_dir = ctx.obj['working_dir']
        outputfile = os.path.join(working_dir, outputfile)
        diagram_ = StkmlDiagram(icons=icons)
        try:
            result = diagram_.diagram_from_stackml(type_=int(type),
                                                   stackml=stackml, output=outputfile)
        except FileNotFoundError as error:
            click.secho(f"[ERROR] {error} ", fg='red')
        except RegionNotFound as error:
            click.secho(f"[ERROR] {error} ", fg='red')
        else:
            if result:
                click.secho(result.rstrip().replace('Warning:', '[WARN]'), fg='yellow')
            click.secho("[SUCCESS] Compilation completed successfully", fg='green')
    else:
        click.secho("""[WARN] failed to execute 'dot',
                    make sure the Graphviz 'https://www.graphviz.org/' executables are on your systems' PATH""",
                    fg='yellow')


@click.command('drawio', short_help='compile Stackml file for Drawio', context_settings=CONTEXT_SETTINGS)
@click.option('--level', '-l', 'level', help='the diagram view level {System View or Layer View}', required=True,
              type=click.Choice(['1', '2']))
@click.option('--icons', '-i', 'icons', default=ICONS, help='the icon folder {it contains [Node_Id].png files}')
@click.option('--output', '-o', 'outputfile', required=True, help='the output file')
@click.pass_context
def drawio(ctx: click.Context, level: str, icons: str, outputfile: str) -> None:
    """
    compile Stackml file for Drawio

    Example:
    \b
        Stackml compile -i architecture.stkml Drawio -l 1 -o Drawio
    """

    stackml = ctx.obj['Stackml']
    working_dir = ctx.obj['working_dir']
    outputfile = os.path.join(working_dir, outputfile)
    template_dir = os.path.join(MODULE_DIR, 'templates')
    env = Environment(loader=FileSystemLoader(template_dir))
    drawio_ = DrawIO(default_icons=ICONS, icons=icons)
    erreur = drawio_.from_stackml(stackml)
    if erreur:
        click.secho(f"[Error] {erreur} ", fg='red')
        sys.exit(1)
    warnings = drawio_.generate_drawio_diagram(output=outputfile, env=env, level=int(level))
    if len(warnings) > 0:
        for warn in warnings:
            click.secho(f"[WARN] {warn} ", fg='yellow')

    click.secho("[SUCCESS] Compilation completed successfully", fg='green')



@click.command('init', help='initialize a Stackml project', context_settings=CONTEXT_SETTINGS)
@click.option('--input', '-i', 'file', help='the file to compile to Stackml', type=str, default=None)
@click.option('--backend', '-b', 'backend', help='the file backend', type=str, default=None)
@click.option('--path', '-p', 'stackmldir', type=str, help='the Stackml project directory', default='.')
def init(stackmldir: str, file: str, backend: str) -> None:
    working_dir = os.path.realpath(os.getcwd())
    #project_dir = os.path.join(working_dir, stackmldir)
    project_dir = check_folder_empty(working_dir, stackmldir)
    if project_dir:
        template_dir = os.path.join(MODULE_DIR, 'templates')
        stackml = Stkml()
        stackml.topology.name = click.prompt('Enter the system Name', type=str)
        stackml.topology.type = '' #click.prompt('Enter the system type', type=str)
        stackml.topology.usecase = click.prompt('Enter the system use case', type=str)
        env = Environment(loader=FileSystemLoader(template_dir))
        if file and backend:
            if backend == 'Drawio':
                file = os.path.join(working_dir, file)
                drawio_ = DrawIO(ICONS, ICONS)
                drawio_.create_drawio(file)
                stackml.from_drawio(drawio_)
        output = stackml.generate_files(environment=env)
        outputfile = os.path.join(project_dir, 'main.stkml.yaml')
        with open(outputfile, 'w') as stackml_file:
            stackml_file.write(output)
        click.secho("[SUCCESS] Stackml project initialized successfully", fg='green')


def check_stackml(stackml, disable_syntax_verification) -> bool:
    try:
        stackml.check_stackml_project()
    except ScannerError as error:
        raise click.ClickException(f'YAML is not valid, {stackml.stackmlfile} {error}')
    else:
        click.secho("[INFO] YAML is valid", fg='blue')
        if disable_syntax_verification:
            click.secho("[INFO] Ignore syntax verification", fg='blue')
        else:
            try:
                stackml.validate_stackml_project()
            except FileNotFoundError as error:
                raise click.ClickException(f'{error} not found')
            except StkmlSyntaxicErrorList as error_list:
                raise click.ClickException(f'Schema validation No\n {error_list}')
            except Exception as error:
                raise click.ClickException(error)
            click.secho("[INFO] Schema validation Ok", fg='blue')

def _check_package(stackml, disable_syntax_verification) -> None:
    try:
        stackml.check_stckml_package()
    except ScannerError as error:
        raise click.ClickException(f'YAML is not valid, {error}')
    else:
        click.secho("[INFO] YAML is valid", fg='blue')
        if disable_syntax_verification:
            click.secho("[INFO] Ignore syntax verification", fg='blue')
        else:
            try:
                stackml.validate_stackml_project()
            except FileNotFoundError as error:
                raise click.ClickException(f'{error} not found')
            except StkmlSyntaxicErrorList as error_list:
                raise click.ClickException(f'Schema validation No\n {error_list}')
            except Exception as error:
                raise click.ClickException(error)
            click.secho("[INFO] Schema validation Ok", fg='blue')

def create_stackml(stackml_folder, ignore_main=False):
    working_dir = os.path.realpath(os.getcwd())
    stackml_folder = os.path.join(working_dir, stackml_folder)
    if not ignore_main:
        if not os.path.isdir(stackml_folder):
            raise click.BadParameter(message=f'{stackml_folder} doesn\'t exist',
                                     param=stackml_folder, param_hint='--input')
        if not os.path.isfile(os.path.join(stackml_folder, 'main.stkml.yaml')):
            raise click.BadParameter(message=f'{os.path.join(stackml_folder,"main.stkml.yaml") } doesn\'t exist',
                                     param=stackml_folder, param_hint='--input')
    stackml = Stkml(stackml_folder)
    return stackml, working_dir

def check_folder_empty(working_dir, outputdir):
    output_dir = os.path.join(working_dir, outputdir)
    if os.path.isdir(output_dir):
        if len(os.listdir(outputdir)) != 0:
            click.secho(f"[WARN] {output_dir} is not empty", fg='yellow')
            if click.confirm('Do you want to delete its content', abort=False):
                shutil.rmtree(output_dir)
                os.makedirs(output_dir)
                return output_dir
            return None
        return output_dir
    if click.confirm(f'{output_dir} Does not existe, do you want to create it', abort=False):
        os.makedirs(output_dir)
        return output_dir
    return None


def which(program):
    return shutil.which(program)

def get_stackml_latest_version() -> str or None:
    try:
        request = requests.get('https://pypi.org/pypi/stkml/json')
    except requests.exceptions.ConnectionError :
        pass
    else:
        version = None
        if request.status_code == 200:
            stackml_pypi = json.loads(request.content.decode("utf-8"))
            version = list(stackml_pypi['releases'].keys())[-1]
        request.close()
        return version

@click.command('check:package', help='check stackml package', context_settings=CONTEXT_SETTINGS,
                hidden=True)
@click.option('--input', '-i', 'package_path', help='the stackml project to check')
@click.option('--disable-syntax-verification', '-d', help='ignore syntax verification', is_flag=True)
def check_package(package_path: str, disable_syntax_verification: bool) -> None:
    stackml, _ = create_stackml(package_path, ignore_main=True)
    try:
        _check_package(stackml, disable_syntax_verification)
    except click.ClickException as error:
        click.secho(f"[ERROR] {error}", fg='red')
        sys.exit(1)
    else:
        click.secho("[SUCCESS] Package checked successfully", fg='green')


cli.add_command(check)
cli.add_command(compile)
cli.add_command(init)
compile.add_command(diagram)
compile.add_command(drawio)
cli.add_command(check_package)
if __name__ == '__main__':
    cli()



#Copyright STACKEO - INRIA 2020 .
class Data:

    def __init__(self, size: float, size_unit: str, type_: str, reference: str):
        self.size = {size: size_unit}
        self.type = type_
        self.reference = reference



#Copyright STACKEO - INRIA 2020 .
# pylint: disable=pointless-statement

from diagrams import Diagram, Cluster, Edge
from Stkml.Diagram.Graphviz.GraphvizManger import GraphvizManger
from Stkml.Diagram.Graphviz import Graphviz
from Stkml.Stkml import Stkml
from Stkml.Diagram.Graphviz.Node import DiagramNode

class SystemDiagram(Graphviz):

    def __init__(self, default_icons: str, icon_dir: str, diagram_attr: dict):
        super().__init__(default_icons=default_icons, icon_dir=icon_dir, diagram_attr=diagram_attr)

    def from_stackml(self, stackml: Stkml, output: str) -> str:
        nodes = {}
        with GraphvizManger() as graphviz_manager:
            with Diagram(stackml.topology.name, filename=output, show=False, graph_attr=self.diagram_attr):
                with Cluster(stackml.topology.name):

                    for tier in stackml.tiers:
                        tier_type = tier.type
                        with Cluster(tier_type):
                            for node in tier.nodes:
                                if self._icon_dir == self.default_icons:
                                    _icon = tier_type
                                else:
                                    _icon = node.id_
                                d_node = DiagramNode(self._icon_dir, _icon, node.id_, quantity=node.cardinality)
                                nodes[node.id_] = d_node
                    sources = []
                    for link in stackml.links:
                        node_source = nodes.get(link.source)
                        node_sink = nodes.get(link.sink)
                        if node_source and node_sink:
                            if node_source not in sources:
                                sources.append(node_source)
                                edge = Edge(color="black")
                                node_source >> edge << node_sink
        return graphviz_manager



#Copyright STACKEO - INRIA 2020
#!/usr/bin/python
import os
MODULE_DIR = os.path.realpath(os.path.join(__file__, '..'))
STACKML_URL_SCHEMA = 'https://stkml.stackeo.io/_static/stackml.schema.json'
STKML_PACKAGES = '.stkml_packages'
STKML_EXTENSION = 'stkml.yaml'
__version__ = '1.0.2a'
